﻿"use strict";

function View(context, canvasXSize, canvasYSize, controller) {
	this.context = context;
	this.canvasXSize = canvasXSize;
	this.canvasYSize = canvasYSize;
	this.controller = controller;
	this.showLoading = false;

	this.roadValues = 16;
	this.roadAngles = new Array(this.roadValues);
	this.resetRoad();
	this.cameraZ = 0;
	this.playerX = 0;
	this.score = 0;
	this.steeringWheelAngle = 0;
	this.shakeSteeringWheel = false;
	this.shakeCounter = 0;
	this.showTitle = false;
	this.showGameOver = false;
	this.ghosts = [];
	this.highestScoreGhost = null;
	this.controlsHelpTimer = 0;
}



View.prototype.load = function() {
	this.showLoading = true;
	var imageFileNames = {
		playButton: "Data/PlayButton.png",
		controlsHelp: "Data/ControlsHelp.png",
		steeringWheel: "Data/SteeringWheel.png",
		fruitLemon: "Data/Lemon.png",
		fruitApple: "Data/Apple.png",
		fruitCherries: "Data/Cherries.png",
		fruitOrange: "Data/Orange.png",
		fruitBanana: "Data/Banana.png",
		obstacleCrate: "Data/Crate.png",
		obstacleGhost: "Data/Ghost.png",
		obstaclePipe: "Data/Pipe.png",
		obstacleRocket: "Data/Rocket.png",
		obstacleBarrel: "Data/Barrel.png",
	};
	this.numImages = 0;
	for (var key in imageFileNames)
		this.numImages++;
	this.images = {};
	for (var key in imageFileNames) {
		var image = new Image();
		image.view = this;
		image.key = key;
		image.onload = function() {
			this.view.images[this.key] = this;
			if (--this.view.numImages === 0)
				this.view.finishedLoading();
		};
		image.src = imageFileNames[key];
	}
};



View.prototype.finishedLoading = function() {
	this.fruitImages = [
		this.images.fruitLemon,
		this.images.fruitApple,
		this.images.fruitCherries,
		this.images.fruitOrange,
		this.images.fruitBanana,
	];
	this.obstacleImages = [
		this.images.obstacleCrate,
		this.images.obstacleGhost,
		this.images.obstaclePipe,
		this.images.obstacleRocket,
		this.images.obstacleBarrel,
	];
	this.showLoading = false;
	this.randomPlayerImage();
	this.controller.finishedLoading();
};



View.prototype.resetRoad = function() {
	for (var i = 0; i < this.roadValues; i++)
		this.roadAngles[i] = 0;
	this.obstacles = new Array();
};



View.prototype.randomPlayerImage = function() {
	this.playerImage = this.fruitImages[randomBelow(this.fruitImages.length)];
};



View.prototype.nextRoad = function(angle) {
	this.roadAngles.shift();
	this.roadAngles.push(angle);
	for (var i = this.obstacles.length - 1; i >= 0; i--) {
		if (this.obstacles[i].z === 0) {
			this.obstacles.splice(i, 1);
		} else {
			this.obstacles[i].z--;
		}
	}
};



View.prototype.isPlayerOffRoad = function() {
	return this.playerX < -17 || this.playerX > 17;
};



var objectSize = 1.4;

View.prototype.hasPlayerHitObstacle = function() {
	for (var i = 0; i < this.obstacles.length; i++) {
		if (this.obstacles[i].z === 0) {
			if (this.playerX + objectSize > this.obstacles[i].x - objectSize &&
					this.playerX - objectSize < this.obstacles[i].x + objectSize)
				return true;
		}
	}
	return false;
};



View.prototype.createObstacle = function() {
	var x = randomBetween(-16, 16);
	var image = this.obstacleImages[randomBelow(this.obstacleImages.length)];
	this.obstacles.push({ x: x, z: this.roadValues - 1, image: image });
};



View.prototype.addPlayerGhost = function(distance) {
	var ghost = { image: this.playerImage, x: this.playerX, z: distance };
	if (this.highestScoreGhost === null) {
		this.highestScoreGhost = ghost;
		return;
	}
	if (this.ghosts.length === 9)
		this.ghosts.shift();
	if (distance > this.highestScoreGhost.distance) {
		this.ghosts.push(this.highestScoreGhost);
		this.highestScoreGhost = ghost;
	} else {
		this.ghosts.push(ghost);
	}
};



function blend(value, min, minValue, max, maxValue) {
	var a = (value - min) / (max - min);
	return (1 - a) * minValue + a * maxValue;
}



View.prototype.render = function() {
	var horizonY = this.projectY(this.roadValues - 1) - 0.5;

	// sky
	var skyGradient = this.context.createLinearGradient(0, 0, 0, horizonY);
	skyGradient.addColorStop(0, "#63C0F2");
	skyGradient.addColorStop(1, "#BADEEF");
	this.context.fillStyle = skyGradient;
	this.context.fillRect(0, 0, this.canvasXSize, horizonY);

	// ground
	this.context.fillStyle = "#58CE3D";
	this.context.fillRect(0, horizonY, this.canvasXSize, this.canvasYSize);

	// road
	var currentAngle = 0;
	var currentX = 0;
	var roadSprites = new Array();
	for (var i = 0; i < this.roadValues - 1; i++) {
		this.context.fillStyle = (this.cameraZ + i) % 5 === 4 ? "#B5B3B3" : "#A8A4A4";
		this.context.strokeStyle = this.context.fillStyle;
		this.context.beginPath();
		var y0 = this.projectY(i);
		var sx0 = this.projectX(currentX - 20, i);
		var ex0 = this.projectX(currentX + 20, i);
		var nextX = currentX + currentAngle + this.roadAngles[i];
		var y1 = this.projectY(i + 1);
		var sx1 = this.projectX(nextX - 20, i + 1);
		var ex1 = this.projectX(nextX + 20, i + 1);
		this.context.moveTo(sx0, y0);
		this.context.lineTo(sx1, y1);
		this.context.lineTo(ex1, y1);
		this.context.lineTo(ex0, y0);
		this.context.closePath();
		this.context.stroke();
		this.context.fill();

		// get info for drawing obstacle sprites
		for (var j = 0; j < this.obstacles.length; j++) {
			var obstacle = this.obstacles[j];
			if (obstacle.z === i) {
				var x = this.projectX(currentX + obstacle.x, i);
				var y = this.projectY(i);
				var size = 40 * this.projectScale(i);
				roadSprites.push({ image: obstacle.image, x: x - size / 2, y: y - size,
						xSize: size, ySize: size });
			}
		}
		// get info for drawing ghost sprites
		if (this.highestScoreGhost !== null && this.highestScoreGhost.z === i + this.cameraZ)
			roadSprites.push(this.createGhostSprite(this.highestScoreGhost, i, currentX));
		for (var j = 0; j < this.ghosts.length; j++) {
			var ghost = this.ghosts[j];
			if (ghost.z === i + this.cameraZ) {
				roadSprites.push(this.createGhostSprite(ghost, i, currentX));
			}
		}

		currentAngle += this.roadAngles[i];
		currentX = nextX;
	}

	// render road sprites
	for (var i = roadSprites.length - 1; i >= 0; i--) {
		var sprite = roadSprites[i];
		if (sprite.opacity !== undefined)
			this.context.globalAlpha = sprite.opacity;
		else
			this.context.globalAlpha = 1;
		this.context.drawImage(sprite.image, sprite.x, sprite.y, sprite.xSize, sprite.ySize);
	}
	this.context.globalAlpha = 1;

	// obstacles
	currentAngle = 0;
	currentX = 0;
	for (var i = 0; i < this.roadValues - 1; i++) {
		currentAngle += this.roadAngles[i];
		currentX += currentAngle;
	}

	if (!this.showLoading) {
		// player sprite
		var projectedPlayerX = this.projectX(this.playerX, 0);
		this.context.drawImage(this.playerImage, projectedPlayerX - 20,
				this.canvasYSize - 12 - 40, 40, 40);
	}

	if (this.showTitle) {
		this.context.font = "bold 58px monospace";
		this.context.fillStyle = "#000000";
		this.drawText("Fruit Convoy", this.canvasXSize / 2, 110, 0.5);

		if (this.showLoading) {
			this.context.font = "bold 22px monospace";
			this.context.fillStyle = "#000000";
			this.drawText("Loading...", this.canvasXSize / 2, 210, 0.5);
		} else {
			// play button
			this.context.drawImage(this.images.playButton, this.canvasXSize / 2 - 90 / 2,
					this.canvasYSize / 2, 90, 50);
		}
	} else if (this.showGameOver) {
		this.context.font = "bold 46px monospace";
		this.context.fillStyle = "#000000";
		this.drawText("Game Over", this.canvasXSize / 2, 80, 0.5);

		this.context.font = "bold 18px monospace";
		this.drawText("Score: " + this.score + "      Highest score: " + this.highestScore,
				this.canvasXSize / 2, 130, 0.5);

		// play button
		this.context.drawImage(this.images.playButton, this.canvasXSize / 2 - 90 / 2,
				this.canvasYSize / 2, 90, 50);
	} else {
		// steering wheel
		this.context.save();
		this.context.translate(this.canvasXSize / 2, 55);
		var angle = this.steeringWheelAngle * 20;
		if (this.shakeSteeringWheel) {
			angle += Math.sin(this.shakeCounter / 4 * (Math.PI * 2)) * 4;
			this.shakeCounter = (this.shakeCounter + 1) % 4;
		}
		this.context.rotate(angle / 360 * (Math.PI * 2));
		this.context.drawImage(this.images.steeringWheel, -25, -25, 50, 50);
		this.context.restore();

		// score
		this.context.font = "16px monospace";
		this.context.fillStyle = "#000000";
		this.drawText("Score: " + this.score, this.canvasXSize - 25, 25, 1);

		// controls help
		if (this.controlsHelpTimer !== 0) {
			--this.controlsHelpTimer;
			this.context.drawImage(this.images.controlsHelp, this.canvasXSize / 2 - Math.floor(131 / 2),
					100, 131, 36);
		}
	}
};



View.prototype.createGhostSprite = function(ghost, i, currentX) {
	var x = this.projectX(currentX + ghost.x, i);
	var y = this.projectY(i);
	if (i === 0)
		y -= 12;
	var size = 40 * this.projectScale(i);
	var opacity = 0.5;
	if (i <= 2)
		opacity = 0;
	else if (i <= 6)
		opacity = blend(i, 1, 0, 6, 0.5)
	return { image: ghost.image, x: x - size / 2, y: y - size, xSize: size, ySize: size,
			opacity: opacity };
};



var n = 2.8;

View.prototype.projectY = function(z) {
	var horizonY = this.canvasYSize / 3.2;
	var groundSize = this.canvasYSize - horizonY;
	return horizonY + groundSize / (z + n) * n;
};



View.prototype.projectX = function(x, z) {
	return this.canvasXSize / 2 + x * 14 / (z + n) * n;
};



View.prototype.projectScale = function(z) {
	return 1 / (z + n) * n;
};



View.prototype.drawText = function(text, x, y, align) {
	var measure = this.context.measureText(text);
	this.context.fillText(text, x - measure.width * align, y);
};